#pragma pack(1)

#define INCL_BASE
#define INCL_NOPM
#define INCL_DOSDEVICES
#include <stdio.h>
#include <os2.h>
#include <malloc.h>

#define DirectoryEntrySize	32

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define	FALSE	0
#endif

char	drive[] = "a:";
typedef	UCHAR	FAT;
typedef FAT		*pFAT;

FAT *allocFAT(PBIOSPARAMETERBLOCK);
void readFAT (pFAT, PBIOSPARAMETERBLOCK, HFILE);
void clearFAT(pFAT, PBIOSPARAMETERBLOCK);
void writeFAT(pFAT, PBIOSPARAMETERBLOCK, HFILE);
void readSector(HFILE, USHORT, pFAT, PBIOSPARAMETERBLOCK);
void writeSector(HFILE, USHORT, pFAT, PBIOSPARAMETERBLOCK);
void scrubDir(HFILE, PBIOSPARAMETERBLOCK);

FSINFO	FSInfo;
USHORT	DiskHasVolumeLabel, OldLabel;

VOID main(int argc, char *argv[])
{
	BYTE	a = 0,
			bCommand = 0;
 	HFILE	hDrive;
	USHORT	usAction,
			rv;
	BIOSPARAMETERBLOCK
			bpb;
	FAT		*pFAT;

	printf("Fast False Floppy Formatter (preliminary) by Jon Saxton, 16 Apr 90\n\n");

	if (argc > 1)
	{
		a = tolower(*argv[1]);
		if (a < 'a' | a > 'b')
			a = 0;
   	}
	if (a == 0)
	{
		puts("Need to specify a (floppy) drive name");
		exit(1);
	}
	drive[0] = a;

	rv = DosQFSInfo(a-'a'+1, 2, (PBYTE)&FSInfo, sizeof FSInfo);
	if (rv & rv != ERROR_NO_VOLUME_LABEL)
	{
		printf("Error %d retrieving volume label", rv);
		exit(1);
	}
	DiskHasVolumeLabel = !rv;
	if (argc > 2)
		if (argv[2][0] == '/' && tolower(argv[2][1]) == 'v')
			if (argc > 3)
			{
				DiskHasVolumeLabel = TRUE;
				OldLabel = FALSE;
				if (strlen(argv[3]) > 11)
					strncpy(FSInfo.vol.szVolLabel,argv[3],11);
				else
					strcpy(FSInfo.vol.szVolLabel,argv[3]);
				FSInfo.vol.cch = strlen(FSInfo.vol.szVolLabel);
			}
			else
				OldLabel = TRUE;
		else
			DiskHasVolumeLabel = FALSE;
	else
		DiskHasVolumeLabel = FALSE;

	if (rv=DosOpen(drive, &hDrive, &usAction, 0L, FILE_NORMAL, FILE_OPEN,
			OPEN_ACCESS_READWRITE |
			OPEN_SHARE_DENYREADWRITE |
			OPEN_FLAGS_DASD, 0L))
	{
		printf("Error %d opening disk %s in DASD mode", rv, drive);
		exit(1);
	}

	DosDevIOCtl(0L, &bCommand, DSK_LOCKDRIVE, IOCTL_DISK, hDrive);
	bCommand=1;
	if (rv=DosDevIOCtl(&bpb, &bCommand,
						DSK_GETDEVICEPARAMS, IOCTL_DISK, hDrive))
	{
		printf("Error %d accessing device parameters for drive %s",
				rv, drive);
		exit(1);
	}

	pFAT = allocFAT(&bpb);
	printf("Reading File Allocation Table\n");
	readFAT (pFAT, &bpb, hDrive);
	printf("Clearing File Allocation Table\n");
	clearFAT(pFAT, &bpb);
	printf("Rewriting File Allocation Table\n");
	writeFAT(pFAT, &bpb, hDrive);
	printf("Scrubbing directory\n");
	scrubDir(hDrive, &bpb);
	if (DiskHasVolumeLabel)
	{
		printf("%sriting volume label\n", OldLabel ? "Rew" : "W");
		bCommand = 0;
		DosDevIOCtl(0L, &bCommand, DSK_REDETERMINEMEDIA, IOCTL_DISK, hDrive);
		DosDevIOCtl(0L, &bCommand, DSK_UNLOCKDRIVE, IOCTL_DISK, hDrive);
		DosClose(hDrive);
		DosSetFSInfo(a-'a'+1, 2, (PBYTE)(&FSInfo.vol), sizeof(FSInfo.vol));
	}
	else
	{
		DosDevIOCtl(0L, &bCommand, DSK_UNLOCKDRIVE, IOCTL_DISK, hDrive);
		DosClose(hDrive);
	}
}

FAT *allocFAT(BIOSPARAMETERBLOCK *bpb)
{
	FAT		*f;

	f = (pFAT)malloc(bpb->usBytesPerSector * bpb->cFATs * bpb->usSectorsPerFAT);
	if (f == NULL)
	{
		printf("Unable to allocate memory for FAT image(s)");
		exit(1);
	}
	return f;
}

void readFAT(FAT *f, BIOSPARAMETERBLOCK *bpb, HFILE hDrive)
{
	USHORT	sector;

	for (sector=0; sector < bpb->usSectorsPerFAT; ++sector)
		readSector(hDrive, sector+1, f+(sector * bpb->usBytesPerSector), bpb);
}

void readSector(HFILE hDrive, USHORT sector, FAT *f, BIOSPARAMETERBLOCK *bpb)
{
	USHORT rv;
	TRACKLAYOUT *tlo;

	tlo = (PTRACKLAYOUT)malloc(sizeof(TRACKLAYOUT)
							   +(bpb->usSectorsPerTrack - 1)*4);

	tlo->bCommand = 0;
	tlo->usHead = (sector % (bpb->usSectorsPerTrack * bpb->cHeads))
				 / bpb->usSectorsPerTrack;
	tlo->usCylinder = sector / (bpb->usSectorsPerTrack * bpb->cHeads);
	tlo->usFirstSector = sector % bpb->usSectorsPerTrack;
	tlo->cSectors = 1;
	for(rv=0; rv<bpb->usSectorsPerTrack; ++rv)
	{
		tlo->TrackTable[rv].usSectorNumber = rv+1;
		tlo->TrackTable[rv].usSectorSize = bpb->usBytesPerSector;
	}

	if (rv=DosDevIOCtl(f, tlo, DSK_READTRACK, IOCTL_DISK, hDrive))
	{
		printf("Error %u reading FAT", rv);
		exit(1);
	}
}

void clearFAT(FAT *f, BIOSPARAMETERBLOCK *bpb)
{
	USHORT	usEntriesInFAT,
			usEntryNumber,
			usEntry,
			FATword,
			*pEntry;

	usEntriesInFAT = bpb->usSectorsPerFAT * bpb->usBytesPerSector * 2 / 3;
	for (usEntryNumber = 2; usEntryNumber < usEntriesInFAT; ++usEntryNumber)
	{
		pEntry = (PUSHORT)(f+(usEntryNumber*3/2));
		FATword = usEntry = *pEntry;
		if (usEntryNumber & 1)
			usEntry >>= 4;
		if ((usEntry &= 0x0FFF) != 0xFF7)
			usEntry = 0;
		if (usEntryNumber & 1)
			FATword = (FATword & 0x000F) | (usEntry << 4);
		else
			FATword = (FATword & 0xF000) | usEntry;
		*pEntry = FATword;
	}
}

void writeFAT(FAT *f, BIOSPARAMETERBLOCK *bpb, HFILE hDrive)
{
	USHORT	sector, secBase;
	BYTE	fat;
	for (fat=0; fat < bpb->cFATs; ++fat)
	{
		secBase = fat * bpb->usSectorsPerFAT + 1;
		for (sector=0; sector < bpb->usSectorsPerFAT; ++sector)
			writeSector(hDrive, sector+secBase,
						f+(sector * bpb->usBytesPerSector), bpb);
	}
}

void writeSector(HFILE hDrive, USHORT sector, FAT *f, BIOSPARAMETERBLOCK *bpb)
{
	USHORT rv;
	TRACKLAYOUT *tlo;

	tlo = (PTRACKLAYOUT)malloc(sizeof(TRACKLAYOUT)
							   +(bpb->usSectorsPerTrack - 1)*4);

	tlo->bCommand = 0;
	tlo->usHead = (sector % (bpb->usSectorsPerTrack * bpb->cHeads))
				 / bpb->usSectorsPerTrack;
	tlo->usCylinder = sector / (bpb->usSectorsPerTrack * bpb->cHeads);
	tlo->usFirstSector = sector % bpb->usSectorsPerTrack;
	tlo->cSectors = 1;
	for(rv=0; rv<bpb->usSectorsPerTrack; ++rv)
	{
		tlo->TrackTable[rv].usSectorNumber = rv+1;
		tlo->TrackTable[rv].usSectorSize = bpb->usBytesPerSector;
	}

	if (rv=DosDevIOCtl(f, tlo, DSK_WRITETRACK, IOCTL_DISK, hDrive))
	{
		printf("Error %u writing FAT", rv);
		exit(1);
	}
}

void scrubDir(HFILE hDrive, BIOSPARAMETERBLOCK *bpb)
{
	BYTE	*secBuf;
	USHORT	sector, secBase;

	secBuf = calloc(1,bpb->usBytesPerSector);
	
	secBase = bpb->cFATs * bpb->usSectorsPerFAT + 1;
	for (sector=0; sector < bpb->cRootEntries / DirectoryEntrySize; ++sector)
		writeSector(hDrive, sector+secBase, secBuf, bpb);
}
