/* touch.c
**
** Copyright (c) 1989, Christopher Laforet
** Released into the Public Domain
**
**
** turbo.c 2.0 version
*/

#define LINT_ARGS


#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <io.h>
#include <dir.h>
#include <dos.h>



#define MAJOR_VERSION		1
#define MINOR_VERSION		0


extern void get_time(unsigned short *time,unsigned short *date);
extern void show_usage(void);
extern void touch(unsigned char *filename,unsigned short time,unsigned short date);

short verbose_flag = 0;
short confirm_flag = 0;
short total_files = 0;
short total_touched = 0;


main(argc,argv)
int argc;
char *argv[];
	{
	unsigned short start_time;
	unsigned short start_date;
	short count;
	short files = 0;

	get_time(&start_time,&start_date);
	fprintf(stderr,"TOUCH (v %u.%02u of %s): A File Date and Time Modification Utility\n",MAJOR_VERSION,MINOR_VERSION,__DATE__);
	fprintf(stderr,"Copyright (c) 1989, Christopher Laforet.  Released into the Public Domain.\n\n");
	if (argc == 1)
		{
		show_usage();
		}
	else
		{
		for (count = 1; count < argc; count++)
			{
			if ((*argv[count] == '/') || (*argv[count] == '-')) 	/* check for flags */
				{
				switch (argv[count][1])
					{
					case '?':
						show_usage();
						break;
					case 'C':
					case 'c':
						confirm_flag = 1;
						break;
					case 'V':
					case 'v':
						verbose_flag = 1;
						break;
					default:
						printf("Warning: Invalid flag argument \"%s\".\n",&argv[count][1]);
						break;
					}
				}
			}
		for (count = 1; count < argc; count++)
			{
			if ((*argv[count] != '/') && (*argv[count] != '-')) 	/* check for files */
				{
				files = 1;
				touch(argv[count],start_time,start_date);
				}
			}
		if (!files)
			{
			printf("Error: No filename(s) specified on command line.\n");
			}
		else
			{
			printf("\nTotal Files Found: %-5u  Total Files Touched: %u\n\n",total_files,total_touched);
			}
		}
	return(0);
	}


void get_time(time,date)
unsigned short *time;
unsigned short *date;
	{
	union REGS registers;

	registers.h.ah = 0x2c;		/* get current time from DOS */
	intdos(&registers,&registers);
	*time = registers.h.ch << 11;			/* hours */
	*time |= (registers.h.cl & 0x3f) << 5;	/* minutes */
	*time |= (registers.h.dh >> 1) & 0x1f;	/* seconds in 2-sec increments */

	registers.h.ah = 0x2a;		/* get current date from DOS */
	intdos(&registers,&registers);
	*date = (registers.x.cx - 1980) << 9;	/* year */
	*date |= (registers.h.dh & 0xf) << 5;	/* month */
	*date |= registers.h.dl & 0xf;			/* day */
	}



void show_usage()
	{
	printf("Usage:  TOUCH [flags] pathname [pathname(s)]\n");
	printf("    where flags are:\n");
	printf("\t-c   Confirm each file to be touched.\n");
	printf("\t-v   Verbose mode.  Show each file touched.\n");
	printf("\t-?   Show this usage screen.\n\n");
	}



void touch(filename,time,date)
unsigned char *filename;
unsigned short time;
unsigned short date;
	{
	unsigned char buffer[100];
	unsigned char drive[MAXDRIVE];
	unsigned char dir[MAXDIR];
	unsigned char file[MAXFILE];
	unsigned char ext[MAXEXT];
	struct ffblk fileblock;
	union REGS registers;
	int fh;
	int rtn;
	short confirm;
	int key;

	fnsplit(filename,drive,dir,file,ext);		/* split pathname */
	rtn = findfirst(filename,&fileblock,FA_ARCH);
	while (!rtn)
		{
		++total_files;
		confirm = 0;
		fnsplit(fileblock.ff_name,buffer,buffer,file,ext);	/* buffer is used as dummy areas */
		fnmerge(buffer,drive,dir,file,ext); 	/* rebuild pathname */
		if (confirm_flag)
			{
			strupr(buffer);
			fprintf(stderr,"Is it OK to touch \"%s\"? ",buffer);
			do
				{
				key = getch();
				key = toupper(key);
				}
			while ((key != 'N') && (key != 'Y'));
			fprintf(stderr,"%c\n",key);
			if (key == 'Y')
				{
				confirm = 1;
				}
			}
		if (!confirm_flag || (confirm_flag && confirm))
			{
			if ((fh = open(buffer,O_RDWR)) == -1)
				{
				printf("Error: Unable to open \"%s\"!\n",buffer);
				}
			else
				{
				if (verbose_flag)
					{
					printf("Touching %-12s -> %02u:%02u:%02u on %02u/%02u/%04u\n",fileblock.ff_name,time >> 11,(time & 0x7e0) >> 5,(time & 0x1f) << 1,(date & 0x1e0) >> 5,date & 0x1f,(date >> 9) + 1980);
					}
				++total_touched;
				registers.x.ax = 0x5701;	 /* set file date and time DOS function */
				registers.x.bx = fh;
				registers.x.cx = time;
				registers.x.dx = date;
				intdos(&registers,&registers);
				if (registers.x.cflag & 1)	/* carry flag set? == DOS error */
					{
					printf("Error: Touch failed on \"%s\"!\n",buffer);
					}
				close(fh);
				}
			}
		rtn = findnext(&fileblock);
		}
	}
